/* (C) 2022 YiRing, Inc. */
package com.yiring.app.service.dept.impl;

import cn.hutool.core.collection.ListUtil;
import cn.hutool.core.convert.Convert;
import cn.hutool.core.util.ObjectUtil;
import cn.hutool.core.util.StrUtil;
import com.github.liaochong.myexcel.core.DefaultStreamExcelBuilder;
import com.github.liaochong.myexcel.core.SaxExcelReader;
import com.yiring.app.excel.dept.DepartmentExportExcel;
import com.yiring.app.excel.dept.DepartmentImportExcel;
import com.yiring.app.param.dept.DepartmentAddParam;
import com.yiring.app.param.dept.DepartmentExportParam;
import com.yiring.app.param.dept.DepartmentFindParam;
import com.yiring.app.param.dept.DepartmentModifyParam;
import com.yiring.app.service.dept.DepartmentService;
import com.yiring.app.vo.dept.DepartmentInfoVo;
import com.yiring.app.vo.dept.DepartmentVo;
import com.yiring.auth.domain.dept.Department;
import com.yiring.auth.domain.dept.DepartmentRepository;
import com.yiring.auth.domain.post.Post;
import com.yiring.auth.param.IdsParam;
import com.yiring.common.core.Result;
import com.yiring.common.core.Status;
import com.yiring.common.param.IdParam;
import com.yiring.common.vo.PageVo;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.concurrent.Executors;
import java.util.stream.Collectors;
import javax.annotation.Resource;
import javax.servlet.http.HttpServletResponse;
import lombok.extern.slf4j.Slf4j;
import org.apache.poi.ss.usermodel.Workbook;
import org.springframework.data.domain.*;
import org.springframework.stereotype.Service;
import org.springframework.transaction.annotation.Transactional;
import org.springframework.web.multipart.MultipartFile;

/**
 * 部门管理
 * @author LJ-2204
 * @date 2022/4/19
 */

@Transactional(rollbackFor = RuntimeException.class)
@Service
@Slf4j
public class DepartmentServiceImpl implements DepartmentService {

    @Resource
    DepartmentRepository departmentRepository;

    @Override
    public Result<String> addDepartment(DepartmentAddParam departmentAddParam) {
        long pidExist = departmentRepository.count(
            Example.of(Department.builder().id(departmentAddParam.getId()).build())
        );

        if (pidExist == 0) return Result.no(Status.BAD_REQUEST, "上级部门不存在");

        long exist = departmentRepository.count(
            Example.of(Department.builder().name(departmentAddParam.getName()).pid(departmentAddParam.getId()).build())
        );
        if (exist > 0) return Result.no(Status.BAD_REQUEST, "同级目录下，已存在该部门");

        Department department = DepartmentAddParam.transform(departmentAddParam);
        departmentRepository.save(department);
        return Result.ok();
    }

    @Override
    public Result<PageVo<DepartmentVo>> findDepartmentTree(DepartmentFindParam departmentFindParam) {
        List<Department> departments = departmentRepository.findAll(
            Example.of(Department.builder().enable(departmentFindParam.getEnable()).build())
        );
        List<DepartmentVo> departmentVos = departments
            .stream()
            .map(DepartmentVo::transform)
            .collect(Collectors.toList());
        List<DepartmentVo> departmentVoList = getChildrenList(0L, null, departmentVos);
        PageVo<DepartmentVo> pageVo = PageVo.build(departmentVoList, departmentVos.size());
        return Result.ok(pageVo);
    }

    @Override
    public Result<DepartmentInfoVo> findDepartmentInfo(IdParam idParam) {
        Optional<Department> departmentOptional = departmentRepository.findOne(
            Example.of(Department.builder().id(idParam.getId()).build())
        );
        if (departmentOptional.isEmpty()) return Result.no(Status.BAD_REQUEST, "部门不存在");
        String pName = null;
        Department department = departmentOptional.get();

        // 渲染上级部门
        if (!ObjectUtil.equals(department.getPid(), 0L)) {
            Optional<Department> department_pid = departmentRepository.findOne(
                Example.of(Department.builder().id(department.getPid()).build())
            );
            if (department_pid.isEmpty()) return Result.no(Status.BAD_REQUEST);
            pName = department_pid.get().getName();
        }
        DepartmentInfoVo departmentInfoVo = DepartmentInfoVo.transform(department);
        departmentInfoVo.setPName(pName);
        return Result.ok(departmentInfoVo);
    }

    @Override
    public void exportDepartment(DepartmentExportParam departmentExportParam, HttpServletResponse response) {
        List<Department> departments = departmentRepository.findAll(
            Example.of(Department.builder().enable(departmentExportParam.getEnable()).build())
        );
        List<DepartmentExportExcel> departmentExportExcels = departments
            .stream()
            .map(DepartmentExportExcel::transform)
            .collect(Collectors.toList());
        try (
            DefaultStreamExcelBuilder<DepartmentExportExcel> streamExcelBuilder = DefaultStreamExcelBuilder
                .of(DepartmentExportExcel.class)
                .threadPool(Executors.newFixedThreadPool(2))
                .rowHeight(14)
                .titleRowHeight(14)
                .style(
                    "cell->vertical-align:center;text-align:center",
                    "title->vertical-align:center;text-align:center;font-weight:bold;font-family:等线"
                )
                .start()
        ) {
            streamExcelBuilder.append(departmentExportExcels);

            String fileName = URLEncoder.encode("部门信息.xlsx", StandardCharsets.UTF_8);
            response.setContentType("application/octet-stream");
            response.setHeader("Access-Control-Expose-Headers", "Content-Disposition");
            response.setHeader("Content-Disposition", "attachment;filename=" + fileName);
            OutputStream out = response.getOutputStream();

            Workbook workbook = streamExcelBuilder.fixedTitles().build();
            workbook.write(out);
            workbook.close();
            out.flush();
            out.close();
        } catch (Exception e) {
            log.error(e.getMessage(), e);
            throw new RuntimeException("导出部门信息失败: " + e.getMessage());
        }
    }

    @Override
    public Result<String> deleteDepartment(IdParam idParam) {
        //todo

        departmentRepository.deleteById(idParam.getId());
        return Result.ok();
    }

    @Override
    public Result<String> stateDepartment(IdParam idParam) {
        // todo
        Optional<Department> departmentOptional = departmentRepository.findOne(
            Example.of(Department.builder().id(idParam.getId()).build())
        );
        if (departmentOptional.isEmpty()) return Result.no(Status.BAD_REQUEST, "部门不存在");

        Department department = departmentOptional.get();
        department.setEnable(!department.getEnable());

        departmentRepository.save(department);
        return Result.ok();
    }

    @Override
    public Result<String> deptBingPost(IdParam idParam, IdsParam idsParam) {
        Optional<Department> departmentOptional = departmentRepository.findOne(
            Example.of(Department.builder().id(idParam.getId()).build())
        );

        if (departmentOptional.isEmpty()) return Result.no(Status.BAD_REQUEST, "部门不存在");

        Department department = departmentOptional.get();

        Long[] ids = Convert.toLongArray(idsParam.getIds().split(","));
        HashSet<Post> posts = new HashSet<>();
        for (Long id : ids) {
            posts.add(Post.builder().id(id).build());
        }

        department.setPosts(posts);

        departmentRepository.save(department);
        return Result.ok();
    }

    @Override
    public Result<String> modifyDept(DepartmentModifyParam departmentModifyParam) {
        Optional<Department> departmentOptional = departmentRepository.findOne(
            Example.of(Department.builder().id(departmentModifyParam.getId()).build())
        );

        if (departmentOptional.isEmpty()) return Result.no(Status.BAD_REQUEST);

        long pidExist = departmentRepository.count(
            Example.of(Department.builder().id(departmentModifyParam.getPid()).build())
        );
        if (pidExist == 0) return Result.no(Status.BAD_REQUEST, "上级部门不存在");

        Department department = departmentOptional.get();
        if (!StrUtil.equals(department.getName(), departmentModifyParam.getName())) {
            long exist = departmentRepository.count(
                Example.of(
                    Department
                        .builder()
                        .name(departmentModifyParam.getName())
                        .pid(departmentModifyParam.getId())
                        .build()
                )
            );
            if (exist > 0) return Result.no(Status.BAD_REQUEST, "同级目录下，已存在该部门");
        }
        Department newDepartment = departmentModifyParam.transform();
        departmentRepository.save(newDepartment);

        return Result.ok();
    }

    @Override
    public Result<String> importDepartment(MultipartFile file) {
        if (file.isEmpty()) return Result.no(Status.BAD_REQUEST, "请选择文件");
        List<DepartmentImportExcel> departmentImportExcels;
        try (InputStream inputStream = file.getInputStream()) {
            departmentImportExcels =
                SaxExcelReader
                    .of(DepartmentImportExcel.class)
                    .rowFilter(row -> row.getRowNum() > 0)
                    .ignoreBlankRow()
                    .read(inputStream);
        } catch (IOException e) {
            log.info(e.getMessage());
            throw new RuntimeException("文件导入异常");
        }

        // 拿到所有部门
        HashMap<String, Department> hashMap = getStringDepartmentHashMap();

        // 对导入信息进行排序
        departmentImportExcels
            .stream()
            .sorted((x, y) -> {
                int xCount = StrUtil.count(x.getName(), "/");
                int yCount = StrUtil.count(y.getName(), "/");
                return Integer.compare(xCount, yCount);
            })
            .forEach(departmentImportExcel -> {
                String name = departmentImportExcel.getName();
                if (ObjectUtil.isNotEmpty(hashMap.get(name))) return;
                int lastIndexOf = name.lastIndexOf("/");
                Department department = DepartmentImportExcel.transform(departmentImportExcel);
                department.setPid(0L);
                if (lastIndexOf == -1) {
                    Department save = departmentRepository.save(department);
                    hashMap.putIfAbsent(name, save);
                } else {
                    String subName = name.substring(0, lastIndexOf);
                    Department hashDept = hashMap.get(subName);
                    department.setName(name.substring(lastIndexOf + 1));
                    department.setPid(hashDept.getId());
                    Department save = departmentRepository.save(department);
                    hashMap.putIfAbsent(name, save);
                }
            });

        return Result.ok();
    }

    public HashMap<String, Department> getStringDepartmentHashMap() {
        List<Department> departments = departmentRepository.findAll();
        HashMap<String, Department> hashMap = new HashMap<>();
        departments.forEach(department -> {
            if (department.getPid() == 0) {
                hashMap.put(department.getName(), department);
            } else {
                ArrayList<String> list = ListUtil.toList();
                List<String> deptNames = getDeptNames(list, department.getPid(), departments);
                ListUtil.reverse(deptNames);
                deptNames.add(department.getName());

                hashMap.put(StrUtil.join("/", deptNames), department);
            }
        });
        return hashMap;
    }

    public List<DepartmentVo> getChildrenList(Long pid, String pName, List<DepartmentVo> departmentVos) {
        return departmentVos
            .stream()
            .filter(departmentVo -> pid.equals(departmentVo.getPid()))
            .peek(departmentVo -> {
                departmentVo.setPName(pName);
                departmentVo.setChildList(getChildrenList(departmentVo.getId(), departmentVo.getName(), departmentVos));
            })
            .collect(Collectors.toList());
    }

    public List<String> getDeptNames(ArrayList<String> list, Long pid, List<Department> departmentList) {
        departmentList
            .stream()
            .filter(department -> pid.equals(department.getId()))
            .forEach(department -> {
                list.add(department.getName());
                getDeptNames(list, department.getPid(), departmentList);
            });
        return list;
    }
}
